/*
 *  DIR.C - based on the DIR internal command of
 *  FreeCOM 0.83 beta 50 (see changelog of cmd/dir.c there).
 */

/* #include "../config.h" */

#include <assert.h>
#include <ctype.h>
/* Unix has no conio, dir, dos, io, but Turbo C has them. */
#include <conio.h>
#include <dir.h>
#include <dos.h>
#include <io.h> /* in Unix: unistd.h */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>

#include <fcntl.h> /* O_RDONLY */

/* String resources */

#include "dirRes.h"

/* not available with Turbo C 2.01 #include <dirent.h> */

/* dfn seems to be from FreeDOS. Used by dir_print_body */
/* requires additional sources: #include <dfn.h> */
#define DFN_DIRECTORY FA_DIREC
/* Other DFN things: char *dfnfilename(char *where) */
/* Other DFN things: char *dfnfullpath(char *where) */
/* Other DFN things: int dfnstat(char *where), returning attribs */

/* old FreeCOM includes: command.h, err_fcts.h, nls.h, strings.h  */
/* Notice that FreeCOM displayString uses int id values, while I  */
/* use char * values for which the whole resource area is parsed. */

#define MEM_ERR printf("Out of memory\n"); return 1;

#define WIDE_COLUMNS 5

/* the boolean command line flags */
static int optO, optS, optP, optW, optB, optA, optL, optY, optV;
char * optOSwitches;

/* Attributes */
static int attrMatch, attrExclude;

int useLFN = 0;

/* *************************************************************** */

/* to make things more self contained: */
#define E_None 0
#define E_Useage 1
#define E_CBreak 3
#define E_NoMem 42
#define E_Other 43
#define E_NotAPathParameter (44)
#define E_FileNotFound (45)

/* wrappers to ease debugging of findfirst / findnext: */
/* #define FINDFIRST(path, attrib, ffblk) findfirst(path, attrib, ffblk) */
/* #define FINDNEXT(ffblk) findnext(ffblk) */
/* maximum text coordinates, taken from BIOS data block: */
#define MAX_X (*(unsigned int  far*)MK_FP(0x40, 0x4a))
#define MAX_Y (*(unsigned char far*)MK_FP(0x40, 0x84))
/* control C handling */
int ctrlBreak = 0;
#define cbreak checkBreak()

/* the filespec used by DIR, among other things */
char *path;
/* current Y position for the /P option */
unsigned line;
/* whether to add newline chars in flush_nl() */
int need_nl;

/****************************************************************************/

void fastStringWrite (char * s, int len) {
#if 0
  if (len <= 0) return;
  fprintf ("%s", s);
#else
  union REGS r;
  if (len <= 0) return;
  r.h.ah = 0x40;
  r.x.dx = (unsigned int)s;
  /* The DS register needs to be set up for non-tiny model */
  r.x.cx = len;
  r.x.bx = 1;		/* fileno(stdout) = 1, by definition */
  int86 (0x21, &r, &r);
#endif
}

int cursColumn = 0;
int cursRow = 1;

void writeString (char *s);

/*
 * increment our line if paginating, display message at end of screen
 */
static int lineIncrement (void) {
  if (!optP)
    return E_None;
  if (++cursRow > MAX_Y) {
    int rv;
    cursRow = 0;
    rv = pause ();
    writeString ("(continuing ");
    writeString (path);
    writeString (")\n");
    return rv;
  }
  return 0;
}

static void writeString (char *s) {
int i, j;
  for (j=0,i=0; ; i++) {
    if ((i-j+cursColumn) >= 80) {
      fastStringWrite (s+j, i-j);
      lineIncrement ();
      j = i;
      cursColumn = 0;
    }
    switch (s[i]) {
      case '\0': {
        if (j >= i) return;
        fastStringWrite (s+j, i-j);
        cursColumn += i - j;
        j = i;
        return;
      }
      case '\n': {
        fastStringWrite (s+j, i-j);
        fastStringWrite ("\r\n", 2);
        lineIncrement ();
        j = i+1;
        cursColumn = 0;
        break;
      }
      default: break;
    }
  }
}

char displayBuffer[384];

/*
 * wrapper for writeString, giving printf functionality
 */
void displayString (char * fmt, ...) {
  va_list ap;
  va_start (ap, fmt);
  vsprintf (displayBuffer, fmt, ap);
  va_end (ap);
  writeString (displayBuffer);
}

/****************************************************************************/

struct squeezedFileBlk {
  unsigned long  sqfb_fileSize;
  unsigned short sqfb_lastModificationTime[2];
  unsigned short sqfb_lastAccessTime;
  unsigned char  shortFileName[14];
  unsigned char  sqfb_attrib;
  unsigned char  longFileName[1];
};

struct LFNFileBlk {
  unsigned long  lfn_attrib;
  unsigned short lfn_creationTime[4];
  unsigned short lfn_lastAccessTime[4];
  unsigned short lfn_lastModificationTime[4];
  unsigned long  lfn_fileSize[2];
  unsigned char  rsvd[8];
  unsigned char  longFileName[260];
  unsigned char  shortFileName[14];
  union {
    struct {
      unsigned short iterationHandle;
      unsigned short unicodeProblems;
    } lfn;
    struct {
      char ff_reserved[21];
    } sfn;
  } xtra;
};

void convertSq2LfnBlk (struct LFNFileBlk * lfn, struct squeezedFileBlk * sqfb) {
  memset (lfn, 0, sizeof (struct LFNFileBlk));
  lfn->lfn_attrib                  = sqfb->sqfb_attrib;
  lfn->lfn_creationTime[0]         = sqfb->sqfb_lastModificationTime[0];
  lfn->lfn_creationTime[1]         = sqfb->sqfb_lastModificationTime[1];
  lfn->lfn_lastAccessTime[0]       = 0;
  lfn->lfn_lastAccessTime[1]       = sqfb->sqfb_lastAccessTime;
  lfn->lfn_lastModificationTime[0] = sqfb->sqfb_lastModificationTime[0];
  lfn->lfn_lastModificationTime[1] = sqfb->sqfb_lastModificationTime[1];
  lfn->lfn_fileSize[1]             = sqfb->sqfb_fileSize;
  strcpy (lfn->shortFileName, sqfb->shortFileName);
  strcpy (lfn->longFileName, sqfb->longFileName);
}

struct squeezedFileBlk * convertLfn2SqBlk (struct LFNFileBlk * lfn) {
struct squeezedFileBlk * sqfb;

  sqfb = (struct squeezedFileBlk *) malloc (sizeof (struct squeezedFileBlk) + strlen (lfn->longFileName) + 1);
  if (sqfb == NULL) return NULL;
  memset (sqfb, 0, sizeof (struct squeezedFileBlk));

  sqfb->sqfb_attrib                   = lfn->lfn_attrib;
  sqfb->sqfb_lastAccessTime           = lfn->lfn_lastAccessTime[1];
  sqfb->sqfb_lastModificationTime[0]  = lfn->lfn_lastModificationTime[0];
  sqfb->sqfb_lastModificationTime[1]  = lfn->lfn_lastModificationTime[1];
  sqfb->sqfb_fileSize                 = lfn->lfn_fileSize[1];
  strcpy (sqfb->shortFileName, lfn->shortFileName);
  strcpy (sqfb->longFileName, lfn->longFileName);

  return sqfb;
}

/*
 02h (2)   file not found
 03h (3)   path not found
 04h (4)   too many open files (no handles available)
 05h (5)   access denied
*/

static struct ffblk SFNblk;

int findFirstFB (char * path, struct LFNFileBlk * lfnBlk, int modeMust, int modeAllow) {
  if (path == NULL || lfnBlk == NULL) return -__LINE__;
  modeAllow &= 0x3f;

  if (useLFN) {
    struct SREGS s;
    union REGS r;

    r.h.ch = modeMust;
    r.h.cl = modeAllow;
    r.x.si = 1;
    r.x.ax = 0x714E; /* long filename - find first matching file */
    r.x.dx = FP_OFF (path);
    s.ds   = FP_SEG (path);
    r.x.di = FP_OFF (lfnBlk);
    s.es   = FP_SEG (lfnBlk);
    r.x.cflag = 1;
    int86x (0x21, &r, &r, &s);

    if (r.x.cflag != 0) return -__LINE__;

    if (lfnBlk->shortFileName[0] == '\0') {
      strncpy (lfnBlk->shortFileName, lfnBlk->longFileName, 13);
    }

    lfnBlk->xtra.lfn.iterationHandle = r.x.ax;
    lfnBlk->xtra.lfn.unicodeProblems = r.x.cx;

  } else {
    int rv;
    if (path == NULL || lfnBlk == NULL) return -__LINE__;
    rv = findfirst (path, &SFNblk, modeMust & 0x3f);
    if (rv != 0) return rv;
    memset (lfnBlk, 0, sizeof (struct LFNFileBlk));
    memcpy (lfnBlk->xtra.sfn.ff_reserved, SFNblk.ff_reserved, 22);
    lfnBlk->lfn_attrib                  = SFNblk.ff_attrib;
    lfnBlk->lfn_creationTime[0]         = SFNblk.ff_ftime;
    lfnBlk->lfn_creationTime[1]         = SFNblk.ff_fdate;
    lfnBlk->lfn_lastAccessTime[0]       = SFNblk.ff_ftime;
    lfnBlk->lfn_lastAccessTime[1]       = SFNblk.ff_fdate;
    lfnBlk->lfn_lastModificationTime[0] = SFNblk.ff_ftime;
    lfnBlk->lfn_lastModificationTime[1] = SFNblk.ff_fdate;
    lfnBlk->lfn_fileSize[1]             = SFNblk.ff_fsize;
    strcpy (lfnBlk->shortFileName, SFNblk.ff_name);
  }

  return 0;
}

int findNextFB (unsigned short iterationHandle, struct LFNFileBlk * lfnBlk) {
  if (lfnBlk == NULL) return -__LINE__;

  if (useLFN) {
    struct SREGS s;
    union REGS r;
    r.x.bx = iterationHandle;
    r.x.si = 1;
    r.x.ax = 0x714F; /* long filename - find next matching file */
    r.x.di = FP_OFF (lfnBlk);
    s.es   = FP_SEG (lfnBlk);
    r.x.cflag = 1;
    int86x (0x21, &r, &r, &s);
    if (r.x.cflag != 0) return -__LINE__;
    if (lfnBlk->shortFileName[0] == '\0') {
      strncpy (lfnBlk->shortFileName, lfnBlk->longFileName, 13);
    }
    lfnBlk->xtra.lfn.iterationHandle = iterationHandle;
    lfnBlk->xtra.lfn.unicodeProblems = r.x.cx;
  } else {
    int rv;
    memcpy (SFNblk.ff_reserved, lfnBlk->xtra.sfn.ff_reserved, 22);
    rv = findnext (&SFNblk);
    if (rv != 0) return rv;
    memset (lfnBlk, 0, sizeof (struct LFNFileBlk));
    memcpy (lfnBlk->xtra.sfn.ff_reserved, SFNblk.ff_reserved, 22);
    lfnBlk->lfn_attrib                  = SFNblk.ff_attrib;
    lfnBlk->lfn_creationTime[0]         = SFNblk.ff_ftime;
    lfnBlk->lfn_creationTime[1]         = SFNblk.ff_fdate;
    lfnBlk->lfn_lastAccessTime[0]       = SFNblk.ff_ftime;
    lfnBlk->lfn_lastAccessTime[1]       = SFNblk.ff_fdate;
    lfnBlk->lfn_lastModificationTime[0] = SFNblk.ff_ftime;
    lfnBlk->lfn_lastModificationTime[1] = SFNblk.ff_fdate;
    lfnBlk->lfn_fileSize[1]             = SFNblk.ff_fsize;
    strcpy (lfnBlk->shortFileName, SFNblk.ff_name);
  }

  return 0;
}

int findCloseFB (unsigned short iterationHandle) {
struct SREGS s;
union REGS r;

  if (useLFN) {
    r.x.bx = iterationHandle;
    r.x.ax = 0x71A1; /* long filename - find close file */
    r.x.cflag = 1;
    int86x (0x21, &r, &r, &s);
    if (r.x.cflag != 0) return -__LINE__;
  }
  return 0;
}

int canonicalizeFileNamePath (char * outfn, char * infn) {
struct SREGS s;
union REGS r;

  r.x.cx = 0x8002; /* leave the subst'ed drive letters the way they are */
  /* (long filename) "truename" - canonicalize filename or path */
  if (useLFN) r.x.ax = 0x7160;
  else r.h.ah = 0x60;
  r.x.si = FP_OFF (infn);
  s.ds   = FP_SEG (infn);
  r.x.di = FP_OFF (outfn);
  s.es   = FP_SEG (outfn);
  int86x (0x21, &r, &r, &s);
  if (r.x.cflag != 0) return -r.x.ax;
  return 0;
}

/****************************************************************************/

/*
 * This is only a simple replacement...
 * DFN things: char *dfnfilename(char *where)
 * returns the point where the filename inside the
 * filespec (possibly with path) in where begins.
 */
char *dfnfilename (char *where) {
  char *p = strrchr(where,'\\');
  if (p) return ++p;
  p = strrchr(where,':');
  if (p) return ++p;
  return where;
} 

/*
 * This is only a simple replacement...
 * DFN things: char *dfnfullpath(char *where)
 * returns a string of "X:\directory\file" format,
 * even when where contained no drivespec / pathspec.
 * Could translate from LFN to SFN here (would also
 * have to process QUOTING in argv[] generation for LFN).
 */
char *dfnfullpath (char *where) {
  char temp[270];
  struct LFNFileBlk lfnBlk;
  int x;

  temp[0] = '\0';
  temp[269] = '\0';
  if (!where) return where;
  /* for (;*where == ' ' || *where == '\t'; where++) {}; */
  temp[0] = 'A';
  temp[1] = ':';
  if (where[1] == ':') {
    temp[0] = toupper (where[0]);
    where += 2;
  } else {
    temp[0] = 'A' + getdisk ();
  }

  if (strchr (where,'\\')) {
    sprintf (temp + 2, "%s", where); /* path was present */
  } else {
    char *cwd;
    temp[2] = '\\';
    temp[3] = 0;
    cwd = &temp[3];
    if (getcurdir (1+temp[0]-'A', cwd) == 0) {
      int len = strlen (cwd);
      if (len > 0) {
        cwd[len] = '\\';
        strcpy (cwd + len + 1, where);
      } else {
        strcat (temp, where);
      }
    } else {
      return 0;
    }
  }

  /* If someone gives a full path, we need to have some way of making a 
     determination if what is given is a directory or single filename */

  for (x = 0; ; x++) {
    if (temp [x] == '?' || temp [x] == '*') {
      x = ~0;
      lfnBlk.lfn_attrib = ~0l;
      break;
    } else if (temp[x] == '\0') {
      x = findFirstFB (temp, &lfnBlk, FA_DIREC, ~FA_LABEL);
      break;
    }
  }

  if (x == 0) {
    if (lfnBlk.lfn_attrib & FA_DIREC) {
      x = strlen (temp);
      switch (temp[x-1]) {
        case '\\': case ':': break;
        case '.': {
          if (x == 1 || (x > 1 && temp[x-1] == '.') 
           || (x > 1 && temp[x-2] == '\\')) break;
	}
        default: {
          strcat (temp, "\\."); break;
        }
      }
    }
    findCloseFB (lfnBlk.xtra.lfn.iterationHandle);
  }

  /* printf ("dfnfullpath[%d] of <%s> is <%s> attrs=[%d:%x]\n", __LINE__, where, temp, x, lfnBlk.lfn_attrib); */
  return strdup (temp);
} 

/*
 * This is only a simple replacement...
 * DFN things: int dfnstat(char *where), returning attribs,
 * or 0 if file not found.
 */
int dfnstat (char *where) {
  struct SREGS s;
  union REGS r;
  r.x.ax = 0x4300; /* get file attributes */
  s.ds = FP_SEG (&where);
  r.x.dx = FP_OFF (&where);
  int86x (0x21, &r, &r, &s);
  return (r.x.cflag == 0) ? r.x.cx : 0;
}

/*
 * translate various drive spec formats into one.
 * Taken from lib/drvnum.c of FreeCOM.
 */
int drvNum (int drive) {
  if (drive == 0)
    return getdisk ();
  if (drive <= 32)
    return drive - 1;
  return toupper (drive) - 'A';
}

/* *************************************************************** */

#define NLS_MAKE_SHORT_AMPM 0

static char * digit4Concat (char * s, int x) {
  s[0] = '0' +  (x / 1000);
  s[1] = '0' + ((x / 100) % 10);
  s[2] = '0' + ((x / 10) % 10);
  s[3] = '0' +  (x % 10);
  return s+4;
}

static char * digit2Concat (char * s, int x) {
  s[0] = '0' + x / 10;
  s[1] = '0' + x % 10;
  return s+2;
}

static char * digitSp2Concat (char * s, int x) {
  s[0] = '0' + x / 10; if (s[0] == '0') s[0] = ' ';
  s[1] = '0' + x % 10;
  return s+2;
}

/*
 * dummy of some NLS aware date / time formatting function,
 * just uses a fixed format: mm-dd-yy or mm-dd-yyyy.
 */
char *nls_makedate (int style, unsigned year, unsigned month, unsigned day) {
  char buf[16];
  char * s;
  buf[0] = '\0';
  if (style);

  s = digit2Concat (buf, month);
  s[0] = '-';
  s = digit2Concat (s+1, day);
  s[0] = '-';
  s = (year > 100) ? digit4Concat (s+1, year) : digit2Concat (s+1, year);
  s[0] = '\0';

  return strdup (buf);
}

/*
 * dummy of some NLS aware date / time formatting function,
 * just uses a fixed format: hh:mm or hh:mm:ss.
 */
char *nls_maketime (int style, unsigned hour, unsigned minute,
  int second, int split) {
  char buf[17];
  char ampm[2];
  char * s;

  if (style == NLS_MAKE_SHORT_AMPM) {
    ampm[1] = '\0';
    ampm[0] = 'a';
    if (hour >= 12) {
      ampm[0] = 'p';
      hour -= 12;
    }
    if (hour == 0) {
      hour = 12;
    }
  }

  buf[0] = '\0';
  if (style || split);

  s = digitSp2Concat (buf, hour);
  s[0] = ':';
  s = digit2Concat (s+1, minute);
  if (second >= 0) {
    s[0] = ':';
    s = digit2Concat (s+1, second);
  }
  if (style == NLS_MAKE_SHORT_AMPM) {
    s[0] = ampm[0];
    s++;
  }

  s[0] = '\0';

  return strdup(buf);
}

/* *************************************************************** */

 void error_out_of_memory (void) {
   if (optB == 0) fprintf (stderr, DS_TEXT_ERROR_OUT_OF_MEMORY);
   /* original version used displayError, by the way */
 }

 void error_file_not_found (void) {
   if (optB == 0) fprintf (stderr, DS_TEXT_ERROR_FILE_NOT_FOUND);
   /* original version used displayError, by the way */
 }

/* *************************************************************** */

/*
 * Control C handling (to make things more self contained)
 */
int checkBreak (void) {
  if (!ctrlBreak)
    return 0;
  ctrlBreak = 0; /* processed */
  return 1;
}

/*
 * insert commas into a number (taken from lib/num_fmt.c
 * of FreeCOM, to make things more self contained)
 */
void convert (unsigned long num, char * const des) {
  char temp[sizeof(num) * 8];
  int c = 0;
  assert(des);
  if (num == 0) { /* represent 0 as "0" */
    des[0] = '0';
    des[1] = 0;
  } else {
    temp[31] = 0;
      do { 
      if (((c + 1) % 4) == 0) temp[30 - c++] = ',';
      temp[30 - c++] = (char) (num % 10) + '0';
    } while ((num /= 10) != 0);
    strcpy(des, &temp[31 - c]);
  }
}

/*
 * pause until a key is pressed, possibly setting ctrlBreak flag.
 * (more self contained than original FreeCOM version)
 */
static int pause (void) {
  int c;
  writeString (DS_TEXT_MSG_PAUSE);

  do {
    c = getch ();
    if (c == '\0')
      c = getch() | 256;
  } while (kbhit ());

  if (c == 3) /* CTRL C */
    ctrlBreak = 1;
  writeString ("\n");
  return 0;
}

static int flush_nl (void) {
  if (need_nl) {
    writeString ("\n");
  }
  return E_None;
}

/*
 * print the header for the dir command
 */
static int dir_print_header (int drive) {
	/* one byte alignment */
#pragma -a-
  struct media_id
  {
    int info_level;
    int serial1;
    int serial2;
    char vol_id[11];
    char file_sys[8];
  }
  media;
	/* standard alignment */
#pragma -a.
  struct LFNFileBlk f;
  struct SREGS s;
  union REGS r;
  int currDisk;
  int rv;
  if (cbreak)
    return E_CBreak; /* aborted by user */
  if (optB)
    return 0; /* bare format, so skip this header stuff */
  currDisk = getdisk();

  /* changed to no longer need changeDrive */
  setdisk(drvNum(drive));
  if (getdisk() != drvNum(drive)) {
    /* print some error: INVALID DRIVE */
    fprintf (stderr, "Invalid drive: %c\n.", 'A' + drvNum(drive));
    /* would better use resource system for the above... */
    setdisk(currDisk);
    return 1;
  }
  /* get the media ID of the drive */
  /*
   Format of disk info:
   Offset  Size    Description     (Table 01766)
   00h    WORD    0000h (info level)
   02h    DWORD   disk serial number (binary)
   06h 11 BYTEs   volume label or "NO NAME    " if none present
   11h  8 BYTEs   (AL=00h only) filesystem type (see #01767)
   */
  r.x.ax = 0x6900;
  r.x.bx = drive;
  s.ds = FP_SEG (&media);
  r.x.dx = FP_OFF (&media);
  int86x (0x21, &r, &r, &s);
  /* print drive info */
  displayString (DS_TEXT_DIR_HDR_VOLUME, drive - 1 + 'A');

  rv = findFirstFB ("\\*.*", &f, FA_LABEL, 0x3f);
  if (rv == 0) {
        /* Added to remove "." from labels which are longer than
               8 characters (as DOS does). */
    char *dotptr = strchr(f.shortFileName, '.');
    if (dotptr != 0)
    	if(strlen(dotptr + 1))
          memmove(dotptr, dotptr + 1, strlen(dotptr));
        else *dotptr = '\0'; /* dot at end of name */
    displayString (DS_TEXT_DIR_HDR_VOLUME_STRING, f.shortFileName);
    findCloseFB (f.xtra.lfn.iterationHandle);
  } else {
    writeString (DS_TEXT_DIR_HDR_VOLUME_NONE);
  }

  setdisk (currDisk);
  /* print the volume serial number if the return was successful */
  if (!r.x.cflag) {
	displayString (DS_TEXT_DIR_HDR_SERIAL_NUMBER
	 , media.serial2, media.serial1);
  }
  /* Added to exactly match DOS's formatting. */
  if (optS) {
      writeString ("\n");
  }
  return 0;
}

/*
 * print_summary: prints dir summary
 */
static int print_summary (unsigned long files, unsigned long bytes) {
  char buffer[32];
  if (optB)
    return 0; /* not done for bare format */
  convert(files, buffer);
  displayString (DS_TEXT_DIR_FTR_FILES, buffer);
  convert(bytes, buffer);
  displayString (DS_TEXT_DIR_FTR_BYTES, buffer);
  need_nl = 1;
  return 0;
}

static int print_total (unsigned long files, unsigned long bytes) {
  int rv;
  if(optB)
    return 0; /* not done in bare format */
  rv = flush_nl();
  if(rv == E_None) {
    writeString (DS_TEXT_DIR_FTR_TOTAL_NUMBER);
    return print_summary(files, bytes);
  }
  return rv;
}

struct fat32extFreeS {
  unsigned short hdrSz;
  unsigned short ver;
  unsigned long sectorsPerCluster;
  unsigned long bytesPerSector;
  unsigned long availableClusters;
  unsigned long totalClusters;
  unsigned long physAvailableSectors;
  unsigned long physTotalSectors;
  unsigned long availableAllocationUnits;
  unsigned long totalAllocationUnits;
  unsigned char rsvd[8];
};

struct fat32extFreeS fat32S;

static int getDiskAttributesFat32 (char driveLetter) {
char driveStr[] = "C:\\";
struct SREGS s;
union REGS r;

  fat32S.hdrSz = 0;
  driveStr[0] = toupper (driveLetter);

  r.x.cx = sizeof (fat32S);
  r.x.ax = 0x7303; /* fat32 - get extended free space on drive */
  r.x.dx = FP_OFF (driveStr);
  s.ds   = FP_SEG (driveStr);
  r.x.di = FP_OFF (&fat32S);
  s.es   = FP_SEG (&fat32S);
  int86x (0x21, &r, &r, &s);
  if (r.x.cflag != 0 || fat32S.hdrSz != sizeof (fat32S)) {
    fat32S.hdrSz = 0;
    return -__LINE__;
  }
  return 0;
}

static int getDiskAttributes (char driveLetter) {
  union REGS r;

  if (getDiskAttributesFat32 (driveLetter) < 0) {
    r.h.ah = 0x36;
    r.h.dl = toupper (driveLetter) - 'A' + 1;
    int86 (0x21, &r, &r);
    fat32S.hdrSz             = 0;
    fat32S.sectorsPerCluster = (unsigned long)r.x.ax;
    fat32S.bytesPerSector    = (unsigned long)r.x.cx;
    fat32S.availableClusters = (unsigned long)r.x.bx;
    fat32S.totalClusters     = (unsigned long)r.x.dx;
  }
  return 0;
}

static unsigned long clusterSizeRoundUp (unsigned long sz) {
unsigned long clusterSz;
  getDiskAttributes (*path);
  clusterSz = fat32S.sectorsPerCluster * fat32S.bytesPerSector;
  return ((int) ((sz + clusterSz - 1) / clusterSz)) * clusterSz;
}

static int clustersToBytes (unsigned long * fxpRes, unsigned long clusters) {
unsigned long lsz, tsz;
int sh;

  if ((int)fat32S.hdrSz < 0 || fxpRes == NULL) return -__LINE__;

  sh = 0;

  /* A contorted way of being able to perform >32 bit multiplications.
   * There are various conditions (mostly disks much larger than exist today) 
   * in which accuracy may be lost. */

  lsz = fat32S.sectorsPerCluster;
  while ((lsz >= 65536L) || ((lsz & 1) == 0)) { lsz >>= 1; sh++; }
  tsz = fat32S.bytesPerSector;
  while ((tsz >= 65536L) || ((tsz & 1) == 0)) { tsz >>= 1; sh++; }
  lsz *= tsz;
  tsz = clusters;
  while ((tsz & 1) == 0) { tsz >>= 1; sh++; }

  while (sh < 11) {
    if (lsz > tsz) lsz >>= 1; else tsz >>= 1;
    sh++;
  }
  if (sh < 12) {
    if (lsz > tsz) lsz = (lsz + 1) >> 1; else tsz = (tsz + 1) >> 1;
    sh++;
  }
  while (sh > 12) {
    if (lsz < tsz) lsz <<= 1; else tsz <<= 1;
    sh--;
  }

  /* This multiplication will fail once disks capable of 16,384 GB capacity 
   * become available (hmm ... that may be about 8 yrs from now.) */

  lsz *= tsz;

  *fxpRes = lsz;
  return 0;
}

static int dir_print_free_fat32 (void) {
char buffer[32];
unsigned long lsz;
int rv;

  rv = clustersToBytes (&lsz, fat32S.availableClusters);
  if (rv) return -__LINE__;

  /* If the result is small enough, just dump out and let the older small
   * number mechanism output the size. */

  if (lsz < (1L << (31 - 12))) return -__LINE__;

  convert (lsz >> 8, buffer);
  sprintf (buffer + strlen(buffer), ".%02u", ((lsz & 0xff) * 100) >> 8);

  displayString (DS_TEXT_DIR_FTR_MBS_FREE, buffer);
  return 0;
}

static int dir_print_free_space (void) {
  char buffer[32];
  unsigned long lsz;

  getDiskAttributes (*path);
  if (dir_print_free_fat32 () < 0) {
    lsz = fat32S.sectorsPerCluster 
        * fat32S.bytesPerSector
        * fat32S.availableClusters;
    convert (lsz, buffer);
    displayString (DS_TEXT_DIR_FTR_BYTES_FREE, buffer);
  }
  return E_None;
}

static int dir_print_free (unsigned long dirs) {
  char buffer[32];
  int rv;

  if(optB)
    return 0; /* not done in bare format */
  /* print number of dirs and bytes free */
  convert (dirs, buffer);
  displayString (DS_TEXT_DIR_FTR_DIRS, buffer);

#if 0
  if (optV) {
    /* Dump out the allocated number of bytes for the files listed */
  }
#endif

  rv = dir_print_free_space ();

  if (optV) {
    int rv2;
    unsigned long lsz;

    getDiskAttributes (*path);
    rv2 = clustersToBytes (&lsz, fat32S.totalClusters);
    if (0 == rv2) {
      if (lsz < (1L << (31 - 12))) {
        lsz = fat32S.sectorsPerCluster 
            * fat32S.bytesPerSector
            * fat32S.totalClusters;
        convert (lsz, buffer);
        displayString (DS_TEXT_DIR_FTR_BYTES_TOTAL, buffer);
      } else {
        convert (lsz >> 8, buffer);
        sprintf (buffer + strlen(buffer), ".%02u", ((lsz & 0xff) * 100) >> 8);
        displayString (DS_TEXT_DIR_FTR_MBS_TOTAL, buffer);
      }
      lsz = ((fat32S.totalClusters - fat32S.availableClusters) * 200 + fat32S.totalClusters) / (2*fat32S.totalClusters);
      sprintf (buffer, "%2d%%", lsz);
      displayString (DS_TEXT_DIR_FTR_IN_USE, buffer);
    }
  }

  return rv;
}

int walkDirectory (char * path, int modeMatch, int modeExclude, int (* cb)(struct LFNFileBlk *, void *), void *parm) {
struct LFNFileBlk file;
int ret, len, cc;
char path2[270];

  if (path == NULL) return -__LINE__;
  if (cb == NULL) return -__LINE__;

  len = strlen (path);
  if (len <= 0) return -__LINE__;
  if (path[len-1] == '\\') {
     sprintf (path2, "%s*.*", path);
     path = path2;
  } else if (path[len-1] == '.') {
     sprintf (path2, "%s\\*.*", path);
     path = path2;
  }

  cc = 0;
  if (findFirstFB (path, &file, modeMatch, ~modeExclude) != 0) return E_FileNotFound;
  do {
    assert (strlen (file.shortFileName) < 13);
    if (cbreak) return E_CBreak;
    if (optL) strlwr (file.shortFileName);
    if ((file.lfn_attrib & modeMatch) == modeMatch) {
      cc++;
      ret = cb (&file, parm); 
      if (ret != E_None) {
        findCloseFB (file.xtra.lfn.iterationHandle);
        return ret;
      }
    }
  } while (findNextFB (file.xtra.lfn.iterationHandle, &file) == 0);
  findCloseFB (file.xtra.lfn.iterationHandle);
  if (cc == 0) return E_FileNotFound;
  return 0;
}

struct dirState {
  unsigned long dircount;
  unsigned long bytecount;
  unsigned long filecount;
  unsigned long count;
  int pathlen;
  char * pattern;
  int mlen, slen;
  union {
    struct squeezedFileBlk ** sqfile;
    char ** name;
  } entry;
};

static int walkBuffer (char * path, int modeMatch, int modeExclude, int (* cb)(struct LFNFileBlk *, void *), void *parm) {
struct LFNFileBlk file;
struct dirState * ds = (struct dirState *) parm;
int i, rv = E_None;

   modeExclude = modeExclude;
   path = path;
   for (i=0; i < ds->slen; i++) {
     if (ds->entry.sqfile[i] != NULL) {
       if (rv == E_None && ((ds->entry.sqfile[i]->sqfb_attrib & modeMatch) == modeMatch)) {
         if (cb != NULL) {
           convertSq2LfnBlk (&file, ds->entry.sqfile[i]);
           rv = cb (&file, parm);
         }
       }
       free (ds->entry.sqfile[i]);
       ds->entry.sqfile[i] = NULL;
     }
   }
   ds->slen = 0;
   if (ds->entry.sqfile) {
     free (ds->entry.sqfile);
     ds->entry.sqfile = NULL;
   }
   return rv;
}

/*****************************************************************************/

static char * stroutLeftJustifyString (char * s, int fieldLength, char * input) {
int i;
  for (i=0; input[i]; i++) s[i] = input[i];
  for (; i < fieldLength; i++) s[i] = ' ';
  return s + i;
}

static char * stroutRightJustifyString (char * s, int fieldLength, char * input) {
int i, len;
  len = strlen (input) - 1;
  for (i = fieldLength - 1; i >= 0 && len >= 0; i--, len--) s[i] = input[len];
  for (; i >= 0; i--) s[i] = ' ';
  return s + fieldLength;
}

static char * stroutConcatString (char * s, char * input) {
int i;
  for (i = 0; input[i]; i++) s[i] = input[i];
  return s + i;
}

/*****************************************************************************/

static char lineBuffer[261 + 80];

int existenceTester (struct LFNFileBlk * file, void * parm) {
int * rv = (int *) parm;

  if (file == NULL) return -__LINE__;
  if (parm == NULL) return -__LINE__;

  return * rv;
}

static int dumpDirOut_Normal (struct LFNFileBlk * file, void * parm) {
  struct dirState * ds = (struct dirState *) parm;
  char buffer[sizeof(long) * 4 + 2], *ext;
  char *p, *s;
  int year, month, day;
  int hour, minute;

  if (file == NULL) return -__LINE__;
  if (parm == NULL) return -__LINE__;

  s = lineBuffer;

  if (file->shortFileName[0] == '.') {
    s = stroutLeftJustifyString (s, 13, file->shortFileName);
  } else {
    ext = strrchr (file->shortFileName, '.');
    if (!ext) {
      ext = "";
    } else {
      *ext = '\0';
      ext++;
    }

    s = stroutLeftJustifyString (s, 8, file->shortFileName);
    s = stroutConcatString (s, " ");
    s = stroutLeftJustifyString (s, 3, ext);
    s = stroutConcatString (s, " ");
  }

  if (file->lfn_attrib & FA_DIREC) {
    s = stroutConcatString (s, DS_TEXT_DIR_LINE_SIZE_DIR);
    ds->dircount++;
  } else {
    convert (file->lfn_fileSize[1], buffer);
    s = stroutRightJustifyString (s, 13, buffer);
    s = stroutConcatString (s, " ");
    ds->bytecount += file->lfn_fileSize[1];
    ds->filecount++;
  }

  /* date and time */
  year = (file->lfn_lastModificationTime[1] >> 9) + 80;
  if (optY)
    year += 1900; /* 4 digit year */
  else
    year %= 100;  /* 2 digit year */
  day = file->lfn_lastModificationTime[1] & 0x001f;
  month = (file->lfn_lastModificationTime[1] >> 5) & 0x000f;
  hour = file->lfn_lastModificationTime[0] >> 5 >> 6;
  minute = (file->lfn_lastModificationTime[0] >> 5) & 0x003f;
  p = nls_makedate (0, year, month, day);

  if (!p) {
    error_out_of_memory ();
    return E_NoMem;
  }

  s = stroutConcatString (s, " ");
  s = stroutConcatString (s, p);

  free (p);
  p = nls_maketime (NLS_MAKE_SHORT_AMPM, hour, minute, -1, 0);

  if (!p) {
    error_out_of_memory ();
    return E_NoMem;
  }

  s = stroutConcatString (s, " ");
  s = stroutConcatString (s, p);

  free (p);

  if (file->longFileName[0] != '\0') {
    s = stroutConcatString (s, " ");
    s = stroutConcatString (s, file->longFileName);
  }

  s = stroutConcatString (s, "\n");
  s[0] = '\0';
  writeString (lineBuffer);

  return E_None;
}

static int dumpDirOut_Verbose (struct LFNFileBlk * file, void * parm) {
  struct dirState * ds = (struct dirState *) parm;
  char buffer[32], *ext;
  char *p, *s;
  int year, month, day;
  int hour, minute;
  unsigned long rndUpSz;

  if (file == NULL) return -__LINE__;
  if (parm == NULL) return -__LINE__;

  s = lineBuffer;

  if (file->shortFileName[0] == '.') {
    s = stroutLeftJustifyString (s, 13, file->shortFileName);
  } else {
    ext = strrchr (file->shortFileName, '.');
    if (!ext) {
      ext = "";
    } else {
      *ext = '\0';
      ext++;
    }

    s = stroutLeftJustifyString (s, 8, file->shortFileName);
    s = stroutConcatString (s, " ");
    s = stroutLeftJustifyString (s, 3, ext);
    s = stroutConcatString (s, " ");
  }

  if (file->lfn_attrib & FA_DIREC) {
    s = stroutConcatString (s, DS_TEXT_DIR_LINE_SIZE_DIR);
    ds->dircount++;
    buffer[0] = '\0';
  } else {
    convert (file->lfn_fileSize[1], buffer);
    s = stroutRightJustifyString (s, 13, buffer);
    s = stroutConcatString (s, " ");
    ds->bytecount += file->lfn_fileSize[1];
    ds->filecount++;
    rndUpSz = clusterSizeRoundUp (file->lfn_fileSize[1]);
    convert (rndUpSz, buffer);
  }
  s = stroutRightJustifyString (s, 13, buffer);
  s = stroutConcatString (s, " ");

  /* modification date and time */
  year = (file->lfn_lastModificationTime[1] >> 9) + 80;
  if (optY)
    year += 1900; /* 4 digit year */
  else
    year %= 100;  /* 2 digit year */
  day = file->lfn_lastModificationTime[1] & 0x001f;
  month = (file->lfn_lastModificationTime[1] >> 5) & 0x000f;
  hour = file->lfn_lastModificationTime[0] >> 5 >> 6;
  minute = (file->lfn_lastModificationTime[0] >> 5) & 0x003f;
  p = nls_makedate (0, year, month, day);
  if (!p) {
    error_out_of_memory ();
    return E_NoMem;
  }

  s = stroutConcatString (s, " ");
  s = stroutConcatString (s, p);

  free (p);
  p = nls_maketime (NLS_MAKE_SHORT_AMPM, hour, minute, -1, 0);
  if (!p) {
    error_out_of_memory ();
    return E_NoMem;
  }

  s = stroutConcatString (s, " ");
  s = stroutConcatString (s, p);

  free (p);

  /* last access date */
  year = (file->lfn_lastAccessTime[1] >> 9) + 80;
  if (optY)
    year += 1900; /* 4 digit year */
  else
    year %= 100;  /* 2 digit year */
  day = file->lfn_lastAccessTime[1] & 0x001f;
  month = (file->lfn_lastAccessTime[1] >> 5) & 0x000f;
  hour = file->lfn_lastAccessTime[0] >> 5 >> 6;
  minute = (file->lfn_lastAccessTime[0] >> 5) & 0x003f;
  p = nls_makedate (0, year, month, day);
  if (!p) {
    error_out_of_memory ();
    return E_NoMem;
  }

  s = stroutConcatString (s, "  ");
  s = stroutConcatString (s, p);

  free (p);

  /* Attribute flags */

  s = stroutConcatString (s, "  ");
  buffer[0] = (file->lfn_attrib & FA_RDONLY) ? 'R' : ' ';
  buffer[1] = (file->lfn_attrib & FA_HIDDEN) ? 'H' : ' ';
  buffer[2] = (file->lfn_attrib & FA_SYSTEM) ? 'S' : ' ';
  buffer[3] = (file->lfn_attrib & FA_LABEL ) ? 'V' : ' ';
  buffer[4] = (file->lfn_attrib & FA_DIREC ) ? 'D' : ' ';
  buffer[5] = (file->lfn_attrib & FA_ARCH  ) ? 'A' : ' ';
  buffer[6] = '\0';
  s = stroutConcatString (s, buffer);

  /* We perform a split here, since there appears to be a bug in TC 2.01's
     vsprintf function not filling in the whole string. */

  s[0] = '\0';
  writeString (lineBuffer);
  s = lineBuffer;

  if (file->longFileName[0] != '\0') {
    s = stroutConcatString (s, "     ");
    s = stroutConcatString (s, file->longFileName);
  }

  s = stroutConcatString (s, "\n");
  s[0] = '\0';
  writeString (lineBuffer);

  return E_None;
}

static int domc = 0;

static int dumpDirOut_MultiColumn (struct LFNFileBlk * file, void * parm) {
  struct dirState * ds = (struct dirState *) parm;
  char buffer[sizeof(file->shortFileName) + 3];
  char buff[16];
  int i;

  if (file == NULL) return -__LINE__;
  if (parm == NULL) return -__LINE__;

  if (file->lfn_attrib & FA_DIREC) {
    sprintf (buffer, "[%s]", file->shortFileName); /* directory */
    ds->dircount++;
  } else {
    strcpy (buffer, file->shortFileName); /* file */
    ds->filecount++;
    ds->bytecount += file->lfn_fileSize[1];
  }

  i = !--ds->count;

  stroutLeftJustifyString (buff, 15, buffer);

  if (domc) writeString ("\n");
  domc = 0;

  if (i) {
    buff[15] = '\0';
    domc = 1;
  } else {
    stroutConcatString (buff+15, " ");
    buff[16] = '\0';
  }

  writeString (buff);

  if (i) {
    ds->count = WIDE_COLUMNS;
  }

  return E_None;
}

static int dumpDirOut_Bare (struct LFNFileBlk * file, void * parm) {
  struct dirState * ds = (struct dirState *) parm;
  char * fn;

  if (file == NULL) return -__LINE__;
  if (parm == NULL) return -__LINE__;

  if (strcmp (file->shortFileName, ".") == 0 ||
      strcmp (file->shortFileName, "..") == 0)
    return 0; /* hide "dot" entries in bare format */
  if (optS)
    writeString (path);

  fn = file->shortFileName;
  if (file->longFileName[0] != '\0') fn = file->longFileName;
  if (optL) strlwr (fn);

  writeString (fn);
  writeString ("\n");

  if (file->lfn_attrib & FA_DIREC) {
    ds->dircount++;
  } else {
    ds->filecount++;
    ds->bytecount += file->lfn_fileSize[1];
  }
  return E_None;
}

int dirBufferFileBlockEntry (struct LFNFileBlk * file, void * parm) {
struct dirState * ds = (struct dirState *) parm;

  if (file == NULL) return -__LINE__;
  if (parm == NULL) return -__LINE__;

  /* if (ds->slen > 32) return E_NoMem; */

  if (optL) strlwr (file->shortFileName);

  if (ds->mlen <= 0) {
    ds->mlen = 16;
    ds->entry.sqfile = (struct squeezedFileBlk **) malloc (ds->mlen * sizeof (struct squeezedFileBlk *));
    if (ds->entry.sqfile == NULL) {
      ds->mlen = 0;
      return E_NoMem;
    }
    ds->slen = 0;
  }

  if (ds->mlen <= ds->slen) {
    int i;
    struct squeezedFileBlk **t;
    for (i=ds->mlen*2; i <= ds->slen; i += i) { /* empty */ };
    t = (struct squeezedFileBlk **) realloc (ds->entry.sqfile, i * sizeof (struct squeezedFileBlk *));
    if (t == NULL) {
      return E_NoMem;
    }
    ds->entry.sqfile = t;
    ds->mlen = i;
  }

  ds->entry.sqfile[ds->slen] = convertLfn2SqBlk (file);
  if (ds->entry.sqfile[ds->slen] == NULL) {
    return E_NoMem;
  }
  ds->slen++;

  return E_None;
}

#define CFBReturn(x) return (((x) ^ retInvMode) - retInvMode)

int compareFileBlocks (struct squeezedFileBlk ** f0, struct squeezedFileBlk ** f1) {
int i, r, retInvMode, retInvModeNext;
struct squeezedFileBlk * file0, * file1;

  if (optOSwitches == NULL || f0 == NULL || f1 == NULL) return -__LINE__;

  file0 = *f0;
  file1 = *f1;

  retInvModeNext = 0;

  for (i=0; optOSwitches[i]; i++) {
    retInvMode = retInvModeNext;
    retInvModeNext = 0;
    switch (toupper (optOSwitches[i])) {
      case '-': {
        retInvModeNext = -1;
        break;
      }
      case 'N': {
        char c0, * ext0 = strrchr (file0->shortFileName, '.');
        char c1, * ext1 = strrchr (file1->shortFileName, '.');
        if (ext0 == NULL) ext0 = file0->shortFileName + strlen (file0->shortFileName);
        if (ext1 == NULL) ext1 = file1->shortFileName + strlen (file1->shortFileName);
        c0 = *ext0;
        c1 = *ext1;
        *ext0 = '\0';
        *ext1 = '\0';
        r = strcmp (file0->shortFileName, file1->shortFileName);
        *ext0 = c0;
        *ext1 = c1;
        if (r) CFBReturn (r);
        break;
      }
      case 'E': {
        char * ext0 = strrchr (file0->shortFileName, '.');
        char * ext1 = strrchr (file1->shortFileName, '.');
        if (ext0 == NULL) ext0 = ""; else ext0++;
        if (ext1 == NULL) ext1 = ""; else ext1++;
        r = strcmp (ext0, ext1);
        if (r) CFBReturn (r);
        break;
      }
      case 'S': {
        if (file0->sqfb_fileSize < file1->sqfb_fileSize) CFBReturn (-1);
        if (file0->sqfb_fileSize > file1->sqfb_fileSize) CFBReturn (1);
        break;
      }
      case 'A': { /* As part of the LFN extensions */
        r = file0->sqfb_lastAccessTime - file1->sqfb_lastAccessTime;
        if (r) CFBReturn (r);
        break;
      }
      case 'D': {
        r = file0->sqfb_lastModificationTime[1] - file1->sqfb_lastModificationTime[1];
        if (r) CFBReturn (r);
        r = file0->sqfb_lastModificationTime[0] - file1->sqfb_lastModificationTime[0];
        if (r) CFBReturn (r);
        break;
      }
      case 'G': {
        if ((file0->sqfb_attrib ^ file1->sqfb_attrib) & FA_DIREC) {
          if (file0->sqfb_attrib & FA_DIREC) CFBReturn (-1);
          CFBReturn (1);
        }
        break;
      }
      case '/': return 0;
    }
  }
  return 0;
}

int dir_list (int pathlen, char *pattern , unsigned long *dcnt
  , unsigned long *fcnt, unsigned long *bcnt);

static int dirSubDirRecurse (struct LFNFileBlk * file, void * parm) {
struct dirState * ds = (struct dirState *) parm;
int rv;

  if (file == NULL) return -__LINE__;
  if (parm == NULL) return -__LINE__;

  rv = E_None;
  if ((file->lfn_attrib & FA_DIREC) != 0 /* is directory */
      && strcmp (file->shortFileName, ".") != 0  /* not cur dir */
      && strcmp (file->shortFileName, "..") != 0) {  /* not parent dir */
    char * fn;

    fn = file->shortFileName;
#if 0
    if (file->longFileName[0] != '\0') {
      fn = file->longFileName;
    }
#endif
    strcpy (&path[ds->pathlen], fn);
    rv = dir_list (ds->pathlen + strlen (fn) + 1, ds->pattern, &ds->dircount, &ds->filecount, &ds->bytecount);
    if (rv == E_FileNotFound) rv = E_None;
  }
  return rv;
}

static int firstHeader = 1;

/*
 * list the files in the directory
 */
static int dir_list (int pathlen, char *pattern , unsigned long *dcnt
  , unsigned long *fcnt, unsigned long *bcnt) {
  int len;
  int rv = E_None;
  char pattern2[270];
  char * orgPath;
  struct dirState ds = { 0, 0, 0, WIDE_COLUMNS };

  ds.pathlen = pathlen;
  ds.pattern = pattern;

  len = strlen (pattern);
  if (len <= 0) return -__LINE__;
  if (pattern[len-1] == '\\') {
     sprintf (pattern2, "%s*.*", pattern);
     pattern = pattern2;
  } else if (pattern[len-1] == '.') {
     if (len == 1) {
       pattern = "*.*";
     } else {
       sprintf (pattern2, "%s\\*.*", pattern);
       pattern = pattern2;
     }
  }

  assert (path);
  assert (pattern);
  assert (pathlen >= 2);   /* at least root */

  /* Search for matching entries */
  path[pathlen - 1] = '\\';
  strcpy (&path[pathlen], pattern);

  orgPath = strdup (path);

  rv = E_Other;
  rv = walkDirectory (orgPath, attrMatch, attrExclude, existenceTester, &rv);
  if (rv == E_Other) rv = E_None;

  if ((firstHeader || rv == E_None) && optB == 0) {
    int rvt;
    /* header for one directory, unless bare... */

    rvt = flush_nl ();

    /* path without superflous '\' at its end */
    if(pathlen == 3)     /* root directory */
      path[pathlen] = '\0';    /* path := path w/o filename */
    else
      path[pathlen - 1] = '\0';

    if (!optS) {
      writeString (" ");
    }

    if (canonicalizeFileNamePath (lineBuffer, path) < 0) {
      displayString (DS_TEXT_DIR_DIRECTORY, path);
    } else {
      displayString (DS_TEXT_DIR_DIRECTORY, lineBuffer);
    }

    if (!optW && optV && rv == E_None) {
      writeString (DS_TEXT_VERBOSE_HEADER);
      writeString ("\n");
    } else rv = rvt;

    writeString ("\n");

  } /* findfirst && not bare */

  firstHeader = 0;

  /* if optB && optS the path with trailing backslash is needed,
   * also for optS below do {} while */
  strcpy (&path[pathlen - 1], "\\");

  if (rv == E_None) {
    int (* walk) (char * path, int modeMatch, int modeExclude, int (* cb)(struct LFNFileBlk *, void *), void *parm);
    walk = walkDirectory;
    if (optO) { /* sorting/buffering */
      ds.slen = 0;
      ds.mlen = 0;
      ds.entry.sqfile = NULL;
      rv = walkDirectory (orgPath, attrMatch, attrExclude, dirBufferFileBlockEntry, &ds);
      if (ds.entry.sqfile) {
        if (rv == E_None) {
	  qsort (ds.entry.sqfile, ds.slen, sizeof (struct squeezedFileBlk *), compareFileBlocks);
          walk = walkBuffer;
        } else if (rv == E_NoMem) {
          /* Walk to just free whatever was allocated */
          walkBuffer (orgPath, 0, 0, NULL, &ds);
          /* Will just drop down as if the /S switch were not there */
        }
      }
    }

    if (optB) { /* bare */
      rv = walk (orgPath, attrMatch, attrExclude, dumpDirOut_Bare, &ds);
    } else if (optW) { /* multicolumn */
      rv = walk (orgPath, attrMatch, attrExclude, dumpDirOut_MultiColumn, &ds);
    } else if (optV) { /* verbose */
      rv = walk (orgPath, attrMatch, attrExclude, dumpDirOut_Verbose, &ds);
    } else {
      rv = walk (orgPath, attrMatch, attrExclude, dumpDirOut_Normal, &ds);
    }

    if (walk == walkBuffer && ds.entry.sqfile) {
        walkBuffer (orgPath, 0, 0, NULL, &ds);
    }
  }

  free (orgPath);

  /* when in multicolumn mode, finish line if needed */
  if (rv == E_None && optW && (ds.count != 0)) {
    writeString ("\n");
  }

  if ((rv == E_None) && (ds.filecount || ds.dircount)) {
    /* The code that was here is now in print_summary */
    rv = print_summary (ds.filecount, ds.bytecount);
  } else if (!optS) {
    /* file not found error message unless when in recursive mode */
    error_file_not_found ();
    rv = E_FileNotFound;
  }

  if ((rv == E_None || rv == E_FileNotFound) && optS) { /* recursive mode, no error */
    /* already set for optB && optS before do {} while above 
     *  path[pathlen - 1] = '\\';
     */
    /* *** FOR ALL subdirectories, DO recursive listing *** */
    ds.pathlen = pathlen;
    strcpy (&path[pathlen], "*.*");
    /* rv = walkDirectory (path, FA_DIREC, dirSubDirRecurse, &ds); */
    rv = walkDirectory (path, FA_DIREC, FA_SYSTEM | FA_HIDDEN, dirSubDirRecurse, &ds);
  } /* recursive mode */

  *dcnt += ds.dircount;
  *fcnt += ds.filecount;
  *bcnt += ds.bytecount;
  return rv;
}

/*
 * call dir_list for the specified filespec, using the FreeDOS
 * DFN (directory function?) helpers... Also add the global
 * headers and footers (e.g. sum of files after recursion).
 */
static int dir_print_body (char *arg, unsigned long *dircount) {
  int rv;
  unsigned long filecount, bytecount;
  char *pattern, *cachedPattern;
  char *p;

  /* Added: ignore arguments that are options, like /S */
  if (arg[0] == '/')
    return E_NotAPathParameter;

  /* Modified to pre-allocate path to 270 bytes so that
   * we don't have to realloc() it later.  That was causing
   * "DIR /S" not to work properly.  The path variable cannot
   * be reallocated once dir_list() is called, because dir_list()
   * is recursive.  This will also help to reduce memory
   * fragmentation.
   */
  if((p = dfnfullpath (arg)) == 0) {
    error_out_of_memory ();
    return E_NoMem;
  }

  if((path = realloc(p, 270 * sizeof (char))) == 0) {
    free(p);
    error_out_of_memory();
    return E_NoMem;
  }
  filecount = bytecount = 0;

  /* print the header */
  if((rv = dir_print_header (toupper (path[0]) - 'A' + 1)) == 0) {
    /* There are some directory specs that are not detected by
     * dfnstat() as they are no part of the filesystem in DOS
     */
    pattern = dfnfilename (path);
    assert (p);
    if(!*pattern || (dfnstat (path) & DFN_DIRECTORY) != 0) {
      pattern = strchr (pattern, '\0');
      if (pattern[-1] != '\\')
        ++pattern;
      rv = dir_list (pattern - path, "*.*", dircount, &filecount
	 , &bytecount);
    } else {
      if ((cachedPattern = strdup (pattern)) == 0) {
	error_out_of_memory ();
	rv = E_NoMem;
      } else {
        rv = dir_list (pattern - path, cachedPattern, dircount
           , &filecount, &bytecount);
        free (cachedPattern);
      }
    }
  }

  /* This has to be enhanced for the /V option. */

  if (optS) {
    rv = print_total (filecount, bytecount);
  }

  if (!optB) {
    if (rv == E_None) {
      rv = dir_print_free (*dircount);
    } else {
      int i;
      for (i=0; i < 17; i++) writeString (" ");
      dir_print_free_space ();
    }
  }

  free(path);
  return rv;
}

/****************************************************************************/

/* Detect MS-Windows */

unsigned int GetMSWin9XVer (void) {
  union REGS r;
  r.x.ax = 0x1600;
  int86 (0x2f, &r, &r);
  if ((r.h.al & 0x7f) == 0) return 0; /* 0x80 is OS/2 apparently */
  return r.h.ah + ((int)r.h.al << 8);
}

#if 0
unsigned long GetDosVer (void) {
  union REGS r;
  r.x.ax = 0x3306;
  int86 (0x21, &r, &r);
  return ((unsigned long) r.x.bx) << 16 + r.x.dx;
}

#define WINNT_DOS_VER (0x3205)

int InWindowsNT (void) {
  return GetMSWin9XVer () == 0 && (GetDosVer () >> 16) == WINNT_DOS_VER;
}
#endif

/****************************************************************************/

/*
 * derived from internal dir command, modified to use argc and argv
 * Original name: cmd_dir(char *rest).
 */

int dirmain (int argc, char **argv) {
  int i, cc;
  int rv = 0; /* return value */
  unsigned long dircount;

  /* initialize options */
  optO = optY = optS = optP = optW = optB = optA = optL = 0;
  optOSwitches = NULL;
  attrMatch = 0;
  attrExclude = FA_HIDDEN | FA_SYSTEM | FA_LABEL;

  useLFN = (GetMSWin9XVer () >= 0x0400);

  line = 0;
  for (i = 0; i < argc; i++) {
    if (argv[i][0] == '/') {
      char * op = &argv[i][1];
      char ch;
      int enableFlag, j;

    ArgOperandStart:;

      enableFlag = 1;
      ch = *op;
      if (ch == '-') {
        enableFlag = 0;
        op++;
        ch = *op;
      }
      if ((ch) && (*op == '-'))
        enableFlag = 0;
      switch (toupper (ch)) {
	case 'O': {
          optO = enableFlag;
          if (op[1] == ':') op++;
          optOSwitches = op+1;

          for (j=1; op[j] != '\0' && op[j] != '/'; j++) {
            switch (toupper (op[j])) {
              case '-': case 'N': case 'E': case 'S': case 'A': case 'D': 
              case 'G': break;
              default: writeString (DS_TEXT_INVALID_SWITCH); 
                       writeString (op+j);
                       writeString ("\n");
                       return rv;
            }
          }

          break;
        }
	case 'S': optS = enableFlag; break;
	case 'P': optP = enableFlag; break;
	case 'W': optW = enableFlag; break;
	case 'B': optB = enableFlag; break;
	case 'A': {
          int negateFlag = 0;
          int * attrVar;
          if (op[1] == ':') op++;

          attrMatch = 0;
          attrExclude = FA_LABEL;

          for (j=1; op[j] != '\0' && op[j] != '/' ; j++) {
            attrVar = &attrMatch;
            if (negateFlag) attrVar = &attrExclude;
            negateFlag = 0;
            switch (toupper (op[j])) {
              case 'R': *attrVar |= FA_RDONLY; break;
              case 'H': *attrVar |= FA_HIDDEN; break;
              case 'S': *attrVar |= FA_SYSTEM; break;
              case 'A': *attrVar |= FA_ARCH;   break;
              case 'D': *attrVar |= FA_DIREC;  break;
              case '-': negateFlag = 1; break;
              default: rv = E_Useage;
            }
          }
          attrMatch &= ~attrExclude;
          optA = enableFlag; break;
        }
	case 'L': optL = enableFlag; break;
	case '4':
	case 'Y': optY = enableFlag; break;
        case 'V': optV = enableFlag; break;
        default: rv = E_Useage;
      }
      if (rv == E_Useage) {
        writeString (DS_TEXT_CMDHELP_DIR);
        return rv;
      }

      for (j=0; op[j]; j++) {
         if (op[j] == '/') {
            op += j + 1;
            goto ArgOperandStart;
         }
      }
    }
    else
    {
      /* file argument, handled below */
    }
  } /* for all argv elements */

  /* for each remaining argument, do a directory listing */
  /* if no arguments (filespecs) given, list . (*.*)...  */
  /* Modified to just use all of argv. */
  dircount = 0;
  cc = 0;
  if (argc > 1) {
    for (i = 1; i < argc; i++) {
      if (argv[i][0] != '/') {
        int rvt;
        cc++;
        rvt = dir_print_body (argv[i], &dircount);
        if (rvt == E_FileNotFound) rvt = E_None;
        rv = rv || rvt;
      }
    }
  }

  if (cc == 0) rv = dir_print_body (".", &dircount);

  return rv;
}

int main (int argc, char **argv) {
char * dirCmd = getenv ("DIRCMD");

  if (dirCmd) {
    int state = 0, i, mlen, slen;
    char ** newargs;
    char * newDirCmd = (char *) malloc (strlen (dirCmd) + 1);
    strcpy (newDirCmd, dirCmd);

    mlen = 4;
    slen = 1;
    newargs = (char **) malloc (mlen * sizeof (char *));
    newargs[0] = argv[0];

    for (i=0; newDirCmd[i]; i++) {
      switch (state) {
        case 0: 
          if (newDirCmd[i] == ' ' || newDirCmd[i] == '\t') continue;
          state = 1;
          if (slen >= mlen) {
            mlen += mlen;
            newargs = (char **) realloc (newargs, mlen * sizeof (char *));
          }
          newargs[slen] = newDirCmd + i;
          slen++;
          break;
        case 1: 
          if (newDirCmd[i] != ' ' && newDirCmd[i] != '\t' && newDirCmd[i] != '\0') continue;
          newDirCmd[i] = '\0';
          state = 0;
      }
    }

    for (i=1; i < argc; i++) {
      if (slen >= mlen) {
        mlen += mlen;
        newargs = (char **) realloc (newargs, mlen * sizeof (char *));
      }
      newargs[slen] = argv[i];
      slen++;
    }

    return dirmain (slen, newargs);
  }
  return dirmain (argc, argv);
}

